#!/usr/bin/env python3
##
## -----------------------------------------------------------------
##    This file is part of WAPT Software Deployment
##    Copyright (C) 2012 - 2020  Tranquil IT https://www.tranquil.it
##    All Rights Reserved.
##
##    WAPT helps systems administrators to efficiently deploy
##    setup, update and configure applications.
## ------------------------------------------------------------------
##

from waptserver.tasks import WaptDB
from waptserver.model import *

class SyncDBConnection():

    def set_host_value(self, uuid, key, value):
        with WaptDB():
            h = HostsGlpiSyncStatus.select().where(HostsGlpiSyncStatus.host_id == uuid).first()
            if not h:
                print("[ERR] Host doesn't exist")
                return False
            if not hasattr(h, key):
                print("[ERR] Key doesn't exist in table")
                return False
            setattr(h, key, value)
            h.save()
            return True

    def select_foreign_host(self, uuid):
        with WaptDB():
            query = Hosts.select().where(Hosts.uuid == uuid)
            return query[0]

    def update_hosts_attributes(self, uuid, attributes, local_matching_attributes=[]):
        with WaptDB():
            if not local_matching_attributes:
                local_matching_attributes = attributes
            for u in uuid:
                host = HostsGlpiSyncStatus.select().where(HostsGlpiSyncStatus.host_id == u)
                if len(host):
                    host = host[0]
                else:
                    continue
                for i in range(len(attributes)):
                    foreign_value = getattr(self.select_foreign_host(u), attributes[i])
                    setattr(host, local_matching_attributes[i], foreign_value)
                host.save()

    def load_host(self, uuid):
        with WaptDB():
            (result, created) = HostsGlpiSyncStatus.get_or_create(host=uuid)
            if not created:
                return created
            self.update_hosts_attributes([result.host_id], ['status_hashes'], ['current_hashes'])
            result.last_upload_successful = False
            result.uploaded_hashes = {}
            result.save()
            return created

    def remove_host(self, uuid):
        r = 0
        with WaptDB():
            query = HostsGlpiSyncStatus.delete().where(HostsGlpiSyncStatus.host == uuid)
            r = query.execute()
        return r > 0

    def reset_host(self, uuid):
        self.remove_host(uuid)
        return self.load_host(uuid)

    def load_current_hashes(self):
        with WaptDB():
            hashes = {}
            query = HostsGlpiSyncStatus.select(HostsGlpiSyncStatus.host, HostsGlpiSyncStatus.current_hashes)
            for q in query:
                hashes[q.host_id] = q.current_hashes
            return hashes

    def load_uploaded_hashes(self):
        with WaptDB():
            hashes = {}
            query = HostsGlpiSyncStatus.select(HostsGlpiSyncStatus.host, HostsGlpiSyncStatus.uploaded_hashes)
            for q in query:
                hashes[q.host_id] = q.uploaded_hashes
            return hashes

    def write_json(self, data, key):
        """
        First level of json object must be host uuid

        data: json object
        key: table key to write in
        """

        for uuid in data:
            self.set_host_value(uuid, key, data[uuid])

    def get_sync_id(self, uuid):
        res = HostsGlpiSyncStatus.select(HostsGlpiSyncStatus.id).where(HostsGlpiSyncStatus.host_id == uuid)
        if not len(res) > 0:
            return -1
        return res[0].id

class HostsDBConnection():

    def get_inventory_query(self):
        return Hosts.select(Hosts.host_info, Hosts.wmi, Hosts.dmi, Hosts.uuid, Hosts.host_metrics).order_by(Hosts.uuid)

    def get_uuid_list(self):
        with WaptDB():
            query = Hosts.select(Hosts.uuid).order_by(Hosts.uuid).dicts()
            return [h['uuid'] for h in query]


class SoftwaresDBConnection():

    def load_db_softwares(self):
        with WaptDB():
            installed_softwares = {}
            query = HostSoftwares.select().dicts()
            for soft in query:
                if not soft['host_id'] in installed_softwares:
                    installed_softwares[soft['host_id']] = []
                installed_softwares[soft['host_id']].append(soft)
            return installed_softwares
