#!/usr/bin/env python3
##
## -----------------------------------------------------------------
##    This file is part of WAPT Software Deployment
##    Copyright (C) 2012 - 2020  Tranquil IT https://www.tranquil.it
##    All Rights Reserved.
##
##    WAPT helps systems administrators to efficiently deploy
##    setup, update and configure applications.
## ------------------------------------------------------------------
##

from peewee import *
from playhouse.postgres_ext import *
import json
import os
import sys
import xml.etree.ElementTree
import time
import copy
import logging
from logging.handlers import RotatingFileHandler

from waptserver.glpi_wapt_inventory import *
from waptserver.glpi_db_connection import *
from waptserver.config import wapt_root_dir, rewrite_config_item
from waptserver.tasks import WaptDB, load_config, tasks_config_file
import waptserver.globalsupdate as gl

logger = logging.getLogger('glpiinventory')
logger.setLevel(logging.DEBUG)
stream_handler = logging.StreamHandler()
stream_handler.setLevel(logging.DEBUG)
stream_handler.setFormatter(logging.Formatter('[%(asctime)s] [Glpi Inventory]: %(levelname)s : %(message)s'))
logger.addHandler(stream_handler)

class Tools():

    def write_json(self, data, file='dump.json'):
        with open(file, 'w', encoding='utf-8') as f:
            f.write(u"{:s}".format(json.dumps(data, ensure_ascii=False)))

    def load_json(self, file):
        data = {}
        with open(file, 'r', encoding='utf-8') as f:
            data = json.loads(f.read())
        return data

    def sort_dict(self, d):
        result = {}
        keys = list(d.keys())
        keys.sort()
        for k in keys:
            result[k] = d[k]
        return result

    def merge_dicts(self, d1, d2):
        """
        Replace d1 values by d2 ones
        d1 key created from d2 if doesn't exist yet
        if types are different, d2 erase d1

        return: (d1, d2)

        """
        if d1 == d2:
            return (d1, d2)
        if type(d2) == str or type(d2) == int or type(d2) != type(d1):
            d1 = d2
            return (d1, d2)
        if type(d2) == dict:
            for k in d2.keys():
                if k not in d1.keys():
                    d1[k] = d2[k]
                    continue
                (d1[k], d2[k]) = self.merge_dicts(d1[k], d2[k])
            return (d1, d2)
        if type(d2) == list:
            added = [d2[i] for i in range(len(d2)) if d2[i] not in d1 ]
            d1 = added + [d1[i] for i in range(len(d1)) if d1[i] or (len(d1) > len(added) + i)]
            return (d1, d2)
        return (d1, d2)

    def get_changed_hashes(self, current_hashes, uploaded_hashes, changed_hashes_buff = {}):
        changed_hashes = changed_hashes_buff.copy()
        for db_key in current_hashes.keys():
            if db_key not in uploaded_hashes.keys():
                changed_hashes[db_key] = current_hashes[db_key]
                continue
            if current_hashes[db_key] != uploaded_hashes[db_key]:
                if type(current_hashes[db_key]) == dict:
                    changed_hashes[db_key] = {}
                    changed_hashes[db_key] = self.get_changed_hashes(current_hashes[db_key], uploaded_hashes[db_key], changed_hashes[db_key])
                else:
                    changed_hashes[db_key] = current_hashes[db_key]
        return changed_hashes


hosts_connection = HostsDBConnection()
softwares_connection = SoftwaresDBConnection()
sync_connection = SyncDBConnection()
tools = Tools()

def update_hosts_list():
    all_uuids = hosts_connection.get_uuid_list()
    all_uuids.sort()
    old_uuids = []
    for u in all_uuids:
        if not sync_connection.load_host(u):
            old_uuids.append(u)
    sync_connection.update_hosts_attributes(old_uuids, ['status_hashes'], ['current_hashes'])

def upload_hosts_inventories(glpi_user, glpi_password, fusioninventory_endpoint, force=False, update_uploaded_hashes_every=25, pause_at_latency=[20.0, 15.0], debug_dir=None):
    """
    Upload hosts whose inventories have changed to Glpi
    Retriving data from HostsGlpiSyncStatus table

    args:
        glpi_user, glpi_password: Username and password for glpi server
        fusioninventory_endpoint: Fusion inventory endpoint on Glpi server to send inventories to
        force: Ignore up to date inventories, send everything
        update_uploaded_hashes_every: Update database every X inventories
        pause_at_latency: Trigger a pause when server's response time is too long
    """
    vlist = ['wmi', 'dmi', 'host_info', 'uuid','host_metrics']
    default = tools.load_json(os.path.join(wapt_root_dir, 'waptserver', 'templates', 'waptdefaultinventory.json'))
    counter = [1, 0]
    conf = {}
    hosts_upload_status = []
    cur_host_status = {}

    logger.debug('Updating from databases...')
    update_hosts_list()
    all_installed_softwares = softwares_connection.load_db_softwares()
    query = hosts_connection.get_inventory_query()
    gl.update('glpi_hosts_upload_status', [])

    current_hashes = tools.sort_dict(sync_connection.load_current_hashes())
    uploaded_hashes = tools.sort_dict(sync_connection.load_uploaded_hashes())
    changed_hashes = tools.get_changed_hashes(current_hashes, uploaded_hashes)

    gl.reload_globals()
    with WaptDB():
        if force:
            changed_hashes = current_hashes.copy()
            logger.info("~> Upload forced")

        counter[1] = len(changed_hashes)
        if counter[1] > 0:
            logger.info(f"-------Uploading {counter[1]} inventories to Glpi-------")
            start_time = time.time()

        i = 0
        for host_data in query:
            conf['glpi_server_status'] = gl.force_get('glpi_inventory_upload_status')
            if conf['glpi_server_status'] == -1:
                sync_connection.write_json(uploaded_hashes, 'uploaded_hashes')
                gl.update('glpi_inventory_upload_status', conf['glpi_server_status'])
                logger.warn("~> Task Killed")
                return counter[1]
            if host_data.uuid not in changed_hashes:
                continue
            request_start = time.time()
            logger.debug(f'Uploading ({counter[0]}/{counter[1]})  [{host_data.uuid}]')

            cur_host_status = {
                    "id": str(counter[0]),
                    "uuid": str(host_data.uuid),
                    "host": host_data.computer_fqdn,
                    "status":"Pending"}

            hosts_upload_status.append(cur_host_status)

            conf['glpi_server_status'] = int((counter[0] / counter[1]) * 100)
            gl.update('glpi_inventory_upload_status', conf['glpi_server_status'])

            sync_connection.set_host_value(host_data.uuid, 'last_upload_successful', False)
            d = copy.deepcopy(default)
            d['sync_id'] = sync_connection.get_sync_id(host_data.uuid)

            qlist = [host_data.wmi, host_data.dmi, host_data.host_info, host_data.uuid, host_data.host_metrics]
            for y in range(len(vlist)):
                if qlist[y]:
                    (d[vlist[y]] ,qlist[y]) = tools.merge_dicts(d[vlist[y]] ,qlist[y])
            if host_data.uuid in all_installed_softwares:
                d['installed_softwares'] = all_installed_softwares[host_data.uuid]

            hinv = HostInventory()
            hinv.create_inventory(d)
            if debug_dir:
                if not os.path.isdir(debug_dir):
                    os.makedirs(debug_dir)
                hinv.write_inventory(os.path.join(debug_dir,d['uuid']+'.xml'))

            r = hinv.upload_to_glpi(glpi_user, glpi_password, fusioninventory_endpoint, quiet=True)
            res_val = []
            if r.content and r.content.startswith(b'<?xml'):
                res_tree = xml.etree.ElementTree.fromstring(r.content)
                for n in res_tree.iter('*'):
                    res_val.append(n.tag)
            if r.status_code != 200 or  (not ('REPLY' in res_val)):
                logger.error(f"Failed to upload [{host_data.uuid}]: {r}")
                hosts_upload_status[-1]['status'] = 'Error'
                gl.update('glpi_hosts_upload_status', hosts_upload_status)
            else:
                if host_data.uuid not in uploaded_hashes:
                    uploaded_hashes[host_data.uuid] = {}
                (uploaded_hashes[host_data.uuid], changed_hashes[host_data.uuid]) = tools.merge_dicts(uploaded_hashes[host_data.uuid], changed_hashes[host_data.uuid])
                sync_connection.set_host_value(host_data.uuid, 'last_upload_successful', True)
                logger.debug("Uploaded")
                hosts_upload_status[-1]['status'] = 'Done'
                gl.update('glpi_hosts_upload_status', hosts_upload_status)

            hinv = None
            logger.debug('-----Took {:.3f}s-----'.format(time.time() - request_start))

            if counter[0] % update_uploaded_hashes_every == 0 and counter[0] != 0:
                sync_connection.write_json(uploaded_hashes, 'uploaded_hashes')
            counter[0] += 1

            if (r.elapsed.total_seconds() > pause_at_latency[1]) and (counter[0] + 1 < counter[1]):
                logger.info(f'{pause_at_latency[0]}s pause triggered due to server latency')
                time.sleep(pause_at_latency[0])

        sync_connection.write_json(uploaded_hashes, 'uploaded_hashes')
        if counter[1] > 0:
            logger.info("--------------Finished in {:.3f}s--------------".format(time.time() - start_time))
        conf['glpi_server_status'] = 100
        gl.update('glpi_inventory_upload_status', conf['glpi_server_status'])
    return counter[1]


if __name__ == '__main__':
    usage = """\
    %prog [-c configfile] [--devel] [action]

    WAPT glpi upload.


    """
    parser = OptionParser(usage=usage, version='glpi_upload_hosts.py ')
    parser.add_option(
        '-c',
        '--config',
        dest='configfile',
        default=waptserver.config.DEFAULT_CONFIG_FILE,
        help='Config file full path (default: %default)')
    parser.add_option('-l', '--loglevel', dest='loglevel', default=None, type='choice',
                      choices=['debug',   'warning', 'info', 'error', 'critical'],
                      metavar='LOGLEVEL', help='Loglevel (default: warning)')
    parser.add_option('-f', '--force', dest='force', default=False, action='store_true',
                      help='Force glpi updat even host is up to date')

    (options, args) = parser.parse_args()

    conf = waptserver.config.load_config(options.configfile)
    logger.info('Using config file %s' % options.configfile)

    logger.info('Load database configuration')
    load_db_config(conf)

    setattr(gl,'conf',conf)
    setattr(gl,'tasks_db_dir',os.path.join(wapt_root_dir, 'db'))
    gl.load_shared_variables()

    if args[0] == 'sync':
        server_timeout = [float(v) for v in conf['glpi_server_pause_timeout'].split(',')]
        upload_hosts_inventories(
            conf['glpi_server_user'], conf['glpi_server_pass'],
            conf['glpi_server_endpoint'],
            update_uploaded_hashes_every=conf['glpi_inventory_update_range'], pause_at_latency=server_timeout, force=options.force, debug_dir = conf.get('glpi_inventory_debug_directory'))
